"""
# Exceptions.

Known exceptions that can occur while using the app.
"""


import typing

# ---------------------------------------
# --------- Rest API ERROR -------------
# ---------------------------------------

# Those errors are returned by the Rest API


class RestAPIError(Exception):
    """Base Rest API error."""

    code: int = 0
    message: str = ""
    data: typing.Optional[dict[str, str]] = None

    def __repr__(self) -> str:
        """Display a string representation of the object."""
        return f"<{self.__class__.__name__}: {self.message}>"

    def __str__(self) -> str:
        """Display a string representation of the error."""
        return str(self.message)

    def __init__(self, *args: object, data: typing.Optional[dict[str, typing.Any]] = None) -> None:
        """Add error data."""
        super().__init__(*args)
        self.data = data
        if data and data.get("error"):  # verify if error is in data and if it is not empty
            if isinstance(data["error"], str):
                self.message = data["error"]
            elif isinstance(data["error"], list) and isinstance(data["error"][0], str):
                self.message = ". ".join(data["error"])
            elif "message" in data["error"]:
                self.message = data["error"]["message"]


class Rest400Error(RestAPIError):
    """Data invalid."""

    code = 400
    message = "Invalid data"


class Rest401Error(RestAPIError):
    """Access token declined."""

    code = 401
    message = "Authentication refused."


class Rest403Error(RestAPIError):
    """Insufficient permission to access the requested data."""

    code = 403
    message = "Permission error."


class Rest404Error(RestAPIError):
    """Path or object does not exist."""

    code = 404
    message = "Data not found."


class Rest405Error(RestAPIError):
    """The method is not allowed is on this path."""

    code = 405
    message = "Method or path not allowed."


class Rest409Error(RestAPIError):
    """There was a conflict due to a duplicate transaction."""

    code = 409
    message = "Conflict or duplicate transaction."


class Rest503Error(RestAPIError):
    """The Rest server is down."""

    code = 503
    message = "The Rest server is unreachable."


RestErrorMap = {
    400: Rest400Error,
    422: Rest400Error,
    401: Rest401Error,
    403: Rest403Error,
    404: Rest404Error,
    405: Rest405Error,
    409: Rest409Error,
    # Server Error
    500: Rest503Error,
    501: Rest503Error,
    502: Rest503Error,
    503: Rest503Error,
}



"""
Helpers for views.

Helpers methods used in app.views.
"""


from dataclasses import dataclass
from typing import Any, Generic, Optional

import pydantic
from fastapi import Request

from sap.beanie.document import DocT

from .exceptions import Validation422Error
from .serializers import SerializerT, WSerializerT
from .utils import Flash, FlashLevel, merge_dict_deep, pydantic_format_errors, unflatten_form_data


@dataclass
class FormValidation(Generic[WSerializerT]):
    """Return the result of data validation for a form serializer."""

    data: dict[str, Any]
    errors: dict[str, Any]
    serializer: Optional[WSerializerT]


async def validate_form(
    request: Request,
    serializer_write_class: type[WSerializerT],
    serializer_read_class: Optional[type[SerializerT]] = None,
    instance: Optional[DocT] = None,
) -> FormValidation[WSerializerT]:
    """Check that a submitted form pass validation."""
    form_data: dict[str, Any] = {}

    if serializer_read_class and instance:
        # Means this is an update. So we first populate existing data
        serializer_read: SerializerT = serializer_read_class.read(instance=instance)
        form_data = serializer_read.dict()

    form_data_received = await request.form()
    form_data = merge_dict_deep(form_data, unflatten_form_data(form_data_received))
    form_errors: dict[str, Any] = {}

    async def run_validation() -> WSerializerT:
        """Run serializer validation."""
        serializer_: WSerializerT = serializer_write_class(**form_data, instance=instance)
        await serializer_.run_async_validators()
        return serializer_

    serializer_write: Optional[WSerializerT] = None

    try:
        serializer_write = await run_validation()
    except pydantic.ValidationError as err:
        form_errors = pydantic_format_errors(err.errors())
        msg = "Les informations soumises ne sont pas valides."
        form_errors["__root__"] = {"msg": msg}
        Flash.add_message(request, msg, level=FlashLevel.ERROR)
    except (AssertionError, Validation422Error) as err:
        form_errors["__root__"] = {"msg": str(err)}
        Flash.add_message(request, str(err), level=FlashLevel.ERROR)

    return FormValidation(serializer=serializer_write, data=form_data, errors=form_errors)



import typing

import passlib.context

import pydantic

crypt_context = passlib.context.CryptContext(schemes=["bcrypt"], deprecated="auto")


class PasswordMixin(pydantic.BaseModel):
    """Password Mixin.

    Inherit on User class to have password field and management utilities.
    """

    hashed_password: typing.Optional[str] = None

    def set_password(self, password: str) -> None:
        """Hash a new password and save to the database."""
        self.hashed_password = crypt_context.hash(password)

    def verify_password(self, password: str) -> bool:
        """Verify if a password matches hash in the database."""
        return crypt_context.verify(password, self.hashed_password)



# mypy: ignore-errors
# pylint: skip-file

import inspect
from copy import copy
from typing import Any, List, Optional, Type

from beanie.odm.fields import ExpressionField, Link, LinkInfo, LinkTypes
from beanie.odm.utils.init import Initializer as Initializer
from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.typing import get_origin


def detect_link(field: ModelField) -> Optional[LinkInfo]:
    """
    It detects link and returns LinkInfo if any found.

    :param field: ModelField
    :return: Optional[LinkInfo]
    """
    if issubclass(field.type_, Link):
        if field.allow_none is True:
            return LinkInfo(
                field=field.name,
                model_class=field.sub_fields[0].type_,  # type: ignore
                link_type=LinkTypes.OPTIONAL_DIRECT,
            )
        return LinkInfo(
            field=field.name,
            model_class=field.sub_fields[0].type_,  # type: ignore
            link_type=LinkTypes.DIRECT,
        )
    if (
        inspect.isclass(get_origin(field.outer_type_))
        and issubclass(get_origin(field.outer_type_), list)  # type: ignore
        and len(field.sub_fields) == 1  # type: ignore
    ):
        internal_field = field.sub_fields[0]  # type: ignore
        if internal_field.type_ == Link:
            if field.allow_none is True:
                return LinkInfo(
                    field=field.name,
                    model_class=internal_field.sub_fields[0].type_,  # type: ignore
                    link_type=LinkTypes.OPTIONAL_LIST,
                )
            return LinkInfo(
                field=field.name,
                model_class=internal_field.sub_fields[0].type_,  # type: ignore
                link_type=LinkTypes.LIST,
            )
    return None


def init_document_fields(cls) -> None:
    """
    Init class fields.

    :return: None
    """
    cls.update_forward_refs()

    def check_nested_links(link_info: LinkInfo, prev_models: List[Type[BaseModel]]) -> None:
        if link_info.model_class in prev_models:
            return
        for k, v in link_info.model_class.__fields__.items():
            nested_link_info = detect_link(v)
            if nested_link_info is None:
                continue

            if link_info.nested_links is None:
                link_info.nested_links = {}
            link_info.nested_links[v.name] = nested_link_info
            new_prev_models = copy(prev_models)
            new_prev_models.append(link_info.model_class)
            check_nested_links(nested_link_info, prev_models=new_prev_models)

    if cls._link_fields is None:
        cls._link_fields = {}
    for k, v in cls.__fields__.items():
        path = v.alias or v.name
        setattr(cls, k, ExpressionField(path))

        link_info = detect_link(v)
        if link_info is not None:
            cls._link_fields[v.name] = link_info
            check_nested_links(link_info, prev_models=[])

    cls._hidden_fields = cls.get_hidden_fields()


Initializer.init_document_fields = staticmethod(init_document_fields)



"""
AMQP.

Establish connection to an AMQP server such as RabbitMQ.
"""
import asyncio

import aioamqp
import aioamqp.channel

from sap.settings import DatabaseParams


class AMQPClient:
    """Set up a connection to the AMQP server."""

    transport: asyncio.Transport
    protocol: aioamqp.AmqpProtocol
    channel: aioamqp.channel.Channel
    db_params: DatabaseParams

    async def connect(self) -> None:
        """
        Establish connect to the AMQP server.

        An `__init__` method can't be a coroutine.
        """
        self.transport, self.protocol = await aioamqp.connect(
            host=self.db_params.host,
            port=self.db_params.port,
            login=self.db_params.username,
            password=self.db_params.password,
            virtualhost=self.db_params.db,
            ssl=self.db_params.protocol.endswith("s"),
        )
        self.channel = await self.protocol.channel()



"""
Link.

Update beanie link with better object mapping.
"""
from __future__ import annotations

from typing import Optional, Type

import bson

import beanie

from .document import DocT

# if TYPE_CHECKING:


# else:
#     from beanie import Link

# T = TypeVar("T")


class DBRef(bson.DBRef):
    """Subclass DBref to add typing for id."""

    id: beanie.PydanticObjectId


class Link(beanie.Link[DocT]):
    """Fix typing issue mypy when querying related fields."""

    # id: beanie.PydanticObjectId
    ref: DBRef
    model_class: Type[DocT]
    doc: Optional[DocT]  # This is the prefetched T document

    def __init__(self, ref: DBRef, model_class: Type[DocT]) -> None:
        """Initialize object."""
        super().__init__(ref=ref, model_class=model_class)
        self.ref = ref
        self.id = ref.id
        self.model_class = model_class
        self.doc = None

    async def fetch(self, fetch_links: bool = False) -> DocT:
        """Overwrite fetch to force missing doc to raise error."""
        return await self.model_class.get_or_404(self.ref.id, with_children=True, fetch_links=fetch_links)


# Link = beanie.Link
# Link.__init__ = MLink.__init__
# Link.fetch = MLink.fetch
# Link = MLink



"""
Utils.

Utilities methods to manage worker tasks.
"""
import typing

import celery
import celery.schedules
from celery.events.state import State
from celery.utils.serialization import strtobool
from celery.worker.control import Panel

from .crons import CronTask

# @contextlib.contextmanager
# def context_timeout(seconds: int) -> typing.Iterator[None]:
#     """
#     Provide a timeout context, to automatically abort
#     background tasks that exceeds time limit.
#     """

#     def raise_timeout(signum: int, frame: typing.Optional[types.FrameType]) -> None:
#         raise TimeoutError

#     # Register a function to raise a TimeoutError on the signal.
#     signal.signal(signal.SIGALRM, raise_timeout)

#     # Schedule the signal to be sent after ``time``.
#     signal.alarm(seconds)

#     try:
#         yield
#     except TimeoutError:
#         raise
#     finally:
#         # Unregister the signal so it won't be triggered
#         # if the timeout is not reached.
#         signal.signal(signal.SIGALRM, signal.SIG_IGN)


def match_amqp_topics(topic_alpha: str, topic_beta: str) -> bool:
    """
    Allow comparison on topics pattern.

    This method makes it easy to know if topic_a is contained in topic_b.
    """
    parts_alpha = topic_alpha.split(".")
    parts_beta = topic_beta.split(".")
    if any(a != b and a != "*" and b != "*" for (a, b) in zip(parts_alpha, parts_beta)):
        return False
    return True


class CeleryBeatTaskParams(typing.TypedDict, total=True):
    """Params to schedule a celery beat task."""

    task: str
    schedule: celery.schedules.crontab
    args: list[typing.Any]
    kwargs: dict[str, typing.Any]
    options: dict[str, str]


def register_tasks_with_celery_beat(
    celery_app: celery.Celery, tasks: list[CronTask], options: dict[str, str]
) -> dict[str, CeleryBeatTaskParams]:
    """Retrieve all task params to set up celery beat."""
    beat_schedule: dict[str, CeleryBeatTaskParams] = {}

    for task in tasks:
        uid_args = "-".join([str(x) for x in task.args])
        uid_kwargs = "-".join([str(x) for x in task.kwargs.values()])
        uid = f"{task.get_name()}:{uid_args}:{uid_kwargs}"
        celery_app.register_task(task)
        beat_schedule[uid] = {
            "task": task.get_name(),
            "schedule": task.schedule,
            "args": task.args,
            "kwargs": task.kwargs,
            "options": options,
        }

    return beat_schedule


@Panel.register(
    type="inspect",
    alias="dump_conf",
    signature="[include_defaults=False]",
    args=[("with_defaults", strtobool)],
)
def conf(state: State, with_defaults: bool = False, **kwargs: typing.Any) -> dict[str, str]:
    """
    Override the default `conf` inspect command to effectively disable it.

    This is to stop sensitive configuration information appearing in e.g. Flower.
    (Celery makes an attempt to remove sensitive information, but it is not foolproof.)
    """
    assert state is not None and with_defaults is not None and kwargs is not None  # silent pylint
    return {"error": "Config inspection has been disabled."}



"""
Auth.

Utils to handle user authentication and permissions.
"""
import base64
import binascii
import time
import typing

import jwt

from fastapi import Cookie, Depends, Request, Response, status
from fastapi.exceptions import HTTPException
from starlette.authentication import AuthCredentials, AuthenticationBackend, AuthenticationError, BaseUser
from starlette.requests import HTTPConnection

from AppMain.settings import AppSettings
from sap.beanie import Document
from sap.beanie.exceptions import Object404Error

UserT = typing.TypeVar("UserT", bound=Document)


class JWTAuth:
    """JWT cookie authentication utils.

    This can be used to login/logout user with persistent sessions.
    Mainly useful for Web Apps. For API, it is better to use Basic Authentication in Headers.
    """

    auth_login_url: typing.ClassVar[str] = "/pages/auth/login/"
    auth_cookie_key: typing.ClassVar[str] = "user_session"
    auth_cookie_expires: typing.ClassVar[int] = 60 * 60 * 12  # expiration = 12 hours
    user_model: type[Document]

    def __init__(self, user_model: type[UserT]) -> None:
        """Initialize the JWT auth helper.

        :param user_model: The User model class.
        """
        super().__init__()
        self.user_model = user_model

    def get_auth_login_url(self) -> str:
        """Retrieve the login url where user are redirect in case of auth failure."""
        return self.auth_login_url

    def get_auth_cookie_key(self) -> str:
        """Retrieve key used to define the authentication cookie."""
        return self.auth_cookie_key

    def get_auth_cookie_expires(self) -> int:
        """Retrieve validity in seconds of the authentication cookie."""
        return self.auth_cookie_expires

    def create_token(self, user: UserT) -> str:
        """Get JWT temporary token."""
        expires = self.get_auth_cookie_expires()
        jwt_data = {"exp": int(time.time()) + expires, "user_id": str(user.id)}
        return jwt.encode(payload=jwt_data, key=AppSettings.CRYPTO_SECRET, algorithm="HS256")

    async def find_user(self, jwt_token: str) -> Document:
        """
        Verify that JWT token is valid.

        :param jwt_token: The lifespan of the token in seconds.
        """

        # Raises: jwt.exceptions.InvalidTokenError => Token has expired or is invalid
        jwt_data = jwt.decode(jwt_token, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        if "user_id" not in jwt_data:
            raise jwt.exceptions.InvalidAudienceError("Cannot read user_id.")

        # Raises: Object404Error => User cannot be found
        return await self.user_model.get_or_404(jwt_data["user_id"])

    async def login(self, response: Response, user: UserT) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.set_cookie(
            key=self.get_auth_cookie_key(),
            value=self.create_token(user=user),
            httponly=True,
        )
        return response

    async def logout(self, response: Response) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.delete_cookie(key=self.get_auth_cookie_key(), httponly=True)
        return response

    def depends(self) -> typing.Any:
        """Provide the authenticated user to views that require it."""

        async def retrieve_user(
            request: Request,
            jwt_token: str = Cookie(default="", alias=self.get_auth_cookie_key()),
        ) -> UserT:
            try:
                return await self.find_user(jwt_token=jwt_token)
            except (Object404Error, jwt.exceptions.InvalidTokenError) as exc:
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                    headers={"Location": self.get_auth_login_url()},
                ) from exc

        return Depends(retrieve_user)


class JWTAuthBackend(AuthenticationBackend, JWTAuth):
    """Starlette Backend to authenticate use through JWT Token in Cookies."""

    async def authenticate(self, conn: HTTPConnection) -> typing.Optional[typing.Tuple["AuthCredentials", "BaseUser"]]:
        """Authenticate the user using Cookies."""
        if self.get_auth_cookie_key() not in conn.cookies:
            return None

        jwt_cookie = conn.cookies[self.get_auth_cookie_key()]

        try:
            jwt_data = jwt.decode(jwt_cookie, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        except jwt.exceptions.InvalidTokenError as exc:
            raise AuthenticationError("Invalid JWT token") from exc

        user = await self.user_model.get_or_404(jwt_data["user_id"])

        return AuthCredentials([user.get_scopes()]), user


class BasicAuthBackend(AuthenticationBackend):
    """Starlette Backend to authenticate use through Basic Token in Header."""

    user_model: type[Document]
    auth_key_attribute: str

    def __init__(self, user_model: type[UserT], auth_key_attribute: str = "auth_key") -> None:
        """Initialize the authentication backend.

        :param user_model: The User model
        :param auth_key_attribute: The authentication key attribute on the User model
        """
        super().__init__()
        self.user_model = user_model
        self.auth_key_attribute = auth_key_attribute

    async def authenticate(self, conn: HTTPConnection) -> typing.Optional[typing.Tuple["AuthCredentials", "BaseUser"]]:
        """Authenticate the user use the Authorization headers."""
        if "Authorization" not in conn.headers:
            return None

        auth = conn.headers["Authorization"]
        scheme, credentials = auth.split()
        if scheme.lower() != "basic":
            return None
        try:
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise AuthenticationError("Invalid basic auth credentials") from exc

        username, _, pwd = decoded.partition(":")

        auth_key = getattr(self.user_model, self.auth_key_attribute)
        user = await self.user_model.find_one_or_404(auth_key == (username or pwd))

        return AuthCredentials(user.get_scopes()), user



"""
Documents.

Override beanie Documents to useful methods.
Most of the methods are inspired from Django behavior on querying data.
"""

from datetime import datetime
from typing import TYPE_CHECKING, Any, Mapping, Optional, Type, TypeVar, Union

from pymongo.client_session import ClientSession

import beanie
import pydantic
from beanie import PydanticObjectId

from .exceptions import Object404Error

if TYPE_CHECKING:
    from beanie.odm.documents import DocType
    from beanie.odm.interfaces.find import DocumentProjectionType


class _DocMeta(pydantic.BaseModel):
    """Meta Data allowing to keep trace of Documents versioning and updates."""

    version: int = 0  # version of the document being imported
    source: Optional[str] = ""  # where the data is coming from: webhook, cron
    created: Optional[datetime] = None  # when the document was first imported
    updated: Optional[datetime] = None  # when the document was last updated
    deleted: Optional[datetime] = None  # when the document was deleted, (deleted document may be retained for logging)


class DocMeta(pydantic.BaseModel):
    """Manage meta data and ensure that it's correctly set."""

    doc_meta: _DocMeta = _DocMeta()

    @pydantic.root_validator
    @classmethod
    def validate_doc_meta(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate doc meta on each model update."""
        doc_meta: _DocMeta = values["doc_meta"]
        doc_meta.updated = datetime.now()
        doc_meta.created = doc_meta.created or doc_meta.updated
        return values


class Document(beanie.Document):
    """Subclass beanie.Document that add handy methods."""

    doc_meta: _DocMeta = _DocMeta()

    @pydantic.root_validator
    @classmethod
    def validate_doc_meta(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate doc meta on each model update."""
        doc_meta: _DocMeta = values["doc_meta"]
        doc_meta.updated = datetime.now()
        doc_meta.created = doc_meta.created or doc_meta.updated
        return values

    @classmethod
    async def get_or_404(
        cls: Type["DocType"],
        document_id: Union[PydanticObjectId, str],
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        **pymongo_kwargs: Any,
    ) -> "DocType":
        """Get document by id or raise 404 error if document does not exist."""
        doc_id = document_id if isinstance(document_id, PydanticObjectId) else PydanticObjectId(document_id)
        result = await super().get(
            document_id=doc_id,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            **pymongo_kwargs,
        )
        if not result:
            raise Object404Error
        return result

    @classmethod
    async def find_one_or_404(
        cls: Type["DocType"],
        *args: Union[Mapping[str, Any], bool],
        projection_model: Optional[Type["DocumentProjectionType"]] = None,
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        **pymongo_kwargs: Any,
    ) -> "DocType":
        """Find document from query or raise 404 error if document does not exist."""
        result: Optional["DocType"] = await super().find_one(
            *args,
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            **pymongo_kwargs,
        )
        if not result:
            raise Object404Error
        return result


DocT = TypeVar("DocT", bound=Document)
ModelT = TypeVar("ModelT", bound=pydantic.BaseModel)



"""
Beanie Client.

Initialize connection to the Mongo Database.
"""
import typing
from dataclasses import dataclass

import pymongo.errors
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase

import beanie

from sap.loggers import logger
from sap.settings import DatabaseParams


@dataclass
class MongoConnection:
    """Define a standard cron task response."""

    client: AsyncIOMotorClient
    database: AsyncIOMotorDatabase


class BeanieClient:
    """Set up a connection to the MongoDB server."""

    connections: typing.ClassVar[dict[str, MongoConnection]] = {}

    @classmethod
    async def get_db_default(cls) -> AsyncIOMotorDatabase:
        """Return the default db connection."""
        return cls.connections["default"].database

    @classmethod
    async def init(
        cls,
        mongo_params: DatabaseParams,
        document_models: list[typing.Union[typing.Type[beanie.Document], typing.Type[beanie.View], str]],
        force: bool = False,
    ) -> None:
        """Open and maintain a connection to the database.

        :force bool: Use it for force a connection initialization
        """

        if "default" in cls.connections and not force:
            database: AsyncIOMotorDatabase = cls.connections["default"].database

            try:
                await database.command("ping")
            except pymongo.errors.ConnectionFailure:
                logger.debug("--> Invalidate existing MongoDB connection")
            else:
                logger.debug("--> Using existing MongoDB connection")
                return

        client = AsyncIOMotorClient(mongo_params.get_dns())
        database = client[mongo_params.db]
        cls.connections["default"] = MongoConnection(client=client, database=database)
        await beanie.init_beanie(database, document_models=document_models, allow_index_dropping=True)
        logger.info("--> Establishing new MongoDB connection")



"""
Airtable Cron Tasks.

A version of cron tasks that store the results on Airtable.
Airtable makes it easy to visualized data.
"""

import json
from datetime import datetime
from typing import ClassVar, Optional

import pyairtable.api
import pyairtable.formulas as pyOps
import pyairtable.utils
from pyairtable.api.table import Table

from .crons import CronResponse, CronStat, CronStorage, FetchStrategy


class AirtableStorage(CronStorage):
    """store cron results in Airtable"""

    run_id: str = ""

    PROJECT_NAME: ClassVar[str]

    TABLE_TASKS: ClassVar[Table]
    TABLE_RUNS: ClassVar[Table]
    TABLE_STATS: ClassVar[Table]

    @classmethod
    def get_env_params(cls) -> tuple[str, str]:
        """Return env name and env id on airtable"""
        raise NotImplementedError

    async def record_task(self) -> None:
        """Register the task meta info on Airtable."""
        env_name, env_id = self.get_env_params()

        query = {"Env": env_name, "Name": self.task_name}
        task_info = self.TABLE_TASKS.first(formula=pyOps.match(query), fields=["Name", "Env"])  # type: ignore
        if task_info:
            self.task_id = task_info["id"]

        microapp = self.task_name.split(".")[1]
        fields = {"Project": self.PROJECT_NAME, "Name": self.task_name, "Microapp": microapp, "Env": [env_id]}
        if self.task_id:
            task_record = self.TABLE_TASKS.update(self.task_id, fields=fields)
        else:
            task_record = self.TABLE_TASKS.create(fields=fields)

        self.task_id = task_record["id"]

    async def record_run_start(self) -> None:
        """Record in the DB that the crontask has started."""
        kwargs = self.task.kwargs
        strategy: Optional[FetchStrategy] = kwargs.get("strategy")
        run = self.TABLE_RUNS.create(
            fields={
                "Task": [self.task_id],
                "Status": "Running",
                "Batch Size": kwargs.get("batch_size", ""),
                "Strategy": strategy.name if strategy else "NONE",
                "Arguments": json.dumps({k: v for k, v in kwargs.items() if k not in ["batch_size", "strategy"]}),
                "Started": pyairtable.utils.datetime_to_iso_str(datetime.now()),
            }
        )
        self.run_id = run["id"]

    async def record_run_end(self, response: CronResponse) -> None:
        """Record in the DB that the crontask has ended."""
        self.TABLE_RUNS.update(
            self.run_id,
            fields={
                "Response": json.dumps(response),
                "Status": "Error" if "error" in response else "Success",
                "Ended": pyairtable.utils.datetime_to_iso_str(datetime.now()),
            },
        )

    async def record_stats(self, stats: list[CronStat]) -> None:
        """Record un the DB stats about data to process by this cron."""
        for stat in stats:
            self.TABLE_STATS.create(fields={"Task": [self.task_id], "Key": stat.name, "Value": stat.value})



"""
# Serializers.

Handle data validation.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, Union

from fastapi import Request
from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, ModelField

from sap.beanie.document import DocT, Document

from .pagination import CursorInfo, PaginatedData

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny

    ExcludeType = Union[AbstractSetIntStr, MappingIntStrAny]


class ObjectSerializer(Generic[DocT], BaseModel):
    """Serialize an object for retrieve or list."""

    @classmethod
    def get_id(cls, instance: DocT) -> str:
        """Return the Mongo ID of the object."""
        return str(instance.id)

    @classmethod
    def get_created(cls, instance: DocT) -> datetime.datetime:
        """Return the user creation date."""
        assert instance.doc_meta.created  # let mypy know that this cannot be null
        return instance.doc_meta.created

    @classmethod
    def get_updated(cls, instance: DocT) -> datetime.datetime:
        """Return the user creation date."""
        assert instance.doc_meta.updated  # let mypy know that this cannot be null
        return instance.doc_meta.updated

    @classmethod
    def _get_instance_data(cls, instance: DocT, exclude: Optional["ExcludeType"] = None) -> dict[str, Any]:
        """Retrieve the serializer value from the instance and getters."""
        data = {}
        exclude = exclude or set()
        for field_name, field in cls.__fields__.items():
            if field_name in exclude:
                continue
            if hasattr(cls, f"get_{field_name}"):
                data[field_name] = getattr(cls, f"get_{field_name}")(instance=instance)
            elif issubclass(field.type_, ObjectSerializer):
                related_object = getattr(instance, field_name, None)
                if field.shape == SHAPE_LIST:
                    data[field_name] = (
                        field.type_.read_list(related_object, exclude=field.field_info.exclude)
                        if related_object
                        else []
                    )
                else:
                    data[field_name] = (
                        field.type_.read(related_object, exclude=field.field_info.exclude) if related_object else None
                    )
            else:
                data[field_name] = getattr(instance, field_name)
        return data

    @classmethod
    def read(cls: type["SerializerT"], instance: DocT, exclude: Optional["ExcludeType"] = None) -> "SerializerT":
        """Serialize a single object instance."""
        return cls(**cls._get_instance_data(instance, exclude=exclude))

    @classmethod
    def read_list(
        cls: type["SerializerT"], instance_list: list[DocT], exclude: Optional["ExcludeType"] = None
    ) -> list["SerializerT"]:
        """Serialize a list of objects."""
        return [cls.read(instance, exclude=exclude) for instance in instance_list]

    @classmethod
    def read_page(
        cls,
        instance_list: list[DocT],
        cursor_info: CursorInfo,
        request: Request,
    ) -> PaginatedData["SerializerT"]:
        """Serialize a list of objects."""

        # TODO: implemented proper cursor pagination, for now fake it till you make it.

        page_next = cursor_info.get_next()
        page_previous = cursor_info.get_previous()
        return PaginatedData(
            count=0,
            next=str(request.url.include_query_params(cursor=page_next)) if page_next else None,
            previous=str(request.url.include_query_params(cursor=page_previous)) if page_previous else None,
            data=cls.read_list(instance_list),
        )


SerializerT = TypeVar("SerializerT", bound=ObjectSerializer[Any])


class WriteObjectSerializer(Generic[DocT], BaseModel):
    """Serialize an object for create or update."""

    instance: Optional[DocT] = None

    async def run_async_validators(self) -> None:
        """Check that data pass DB validation."""

        field: ModelField
        embedded_serializers = {}
        for field_name, field in self.__fields__.items():
            if issubclass(field.type_, WriteObjectSerializer):
                embedded_serializers[field_name] = field

        field_serializer: WriteObjectSerializer[DocT]
        for field_name in embedded_serializers:
            if field_serializer := getattr(self, field_name):
                if self.instance:
                    field_serializer.instance = getattr(self.instance, field_name)
                await field_serializer.run_async_validators()

    def dict(
        self,
        *,
        include: Optional["ExcludeType"] = None,
        exclude: Optional["ExcludeType"] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":
        """Dump the serializer data with exclusion of unwanted fields."""
        # Exclude from dumping
        exclude_: dict[Union[int, str], bool] = {"instance": True}
        if exclude:
            for x in exclude:
                exclude_[x] = True

        # Some fields are only excluded from being cascade dumps to dict,
        # but their original value is still needed
        exclude_doc_dumps = {}

        # Embedded documents need to be converted to object after dumps
        embedded_serializers = {}

        for field_name, field in self.__fields__.items():
            if field_name in exclude_:
                continue

            if issubclass(field.type_, Document):
                exclude_doc_dumps[field_name] = True
                exclude_[field_name] = True

            elif issubclass(field.type_, WriteObjectSerializer):
                embedded_serializers[field_name] = field.type_.__fields__["instance"].type_

            # # Some fields are excluded as they are only needed for create
            # if field.field_info.extra.get("exclude_update") and self.instance:
            #     exclude_[field_name] = True

            # # Some fields are excluded as they are only needed for update
            # elif field.field_info.extra.get("exclude_create") and not self.instance:
            #     exclude_[field_name] = True

        result = super().dict(
            include=include,
            exclude=exclude_,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        for field_name in exclude_doc_dumps:
            result[field_name] = getattr(self, field_name)

        instance_embedded: Optional[BaseModel]
        for field_name, field_model in embedded_serializers.items():
            if not result[field_name]:
                continue
            if instance_embedded := getattr(self.instance, field_name, None):
                result[field_name] = instance_embedded.copy(update=result[field_name])
            else:
                result[field_name] = field_model(**result[field_name])

        return result

    async def create(self, **kwargs: Any) -> DocT:
        """Create the object in the database using the data extracted by the serializer."""
        instance_class: type[DocT] = self.__fields__["instance"].type_
        self.instance = await instance_class(**self.dict()).create()
        return self.instance

    async def update(self, **kwargs: Any) -> DocT:
        """Update the object in the database using the data extracted by the serializer."""
        assert self.instance
        instance: DocT = self.instance.copy(update=self.dict())
        await instance.save()
        self.instance = instance
        return instance


WSerializerT = TypeVar("WSerializerT", bound=WriteObjectSerializer[Any])
